@@ -1,4 +1,4 @@
-/* 
+/*
    Copyright (C) 2002, 2003, 2004, 2005, 2006 Bastien Nocera
    Copyright (C) 2003, 2004 Colin Walters <walters@rhythmbox.org>
 
@@ -27,7 +27,8 @@
  * @include: totem-pl-parser.h
  *
  * #TotemPlParser is a general-purpose playlist parser and writer, with
- * support for several different types of playlist.
+ * support for several different types of playlist. Note that totem-pl-parser requires a main loop
+ * to operate properly (e.g. for the #TotemPlParser::entry-parsed signal to be emitted).
  *
  * <example>
  *  <title>Reading a Playlist</title>
@@ -43,6 +44,29 @@
  * g_object_unref (pl);
  *  </programlisting>
  * </example>
+ *
+  * <example>
+ *  <title>Reading a Playlist Asynchronously</title>
+ *  <programlisting>
+ * TotemPlParser *pl = totem_pl_parser_new ();
+ * g_object_set (pl, "recurse", FALSE, "disable-unsafe", TRUE, NULL);
+ * g_signal_connect (G_OBJECT (pl), "playlist-started", G_CALLBACK (playlist_started), NULL);
+ * g_signal_connect (G_OBJECT (pl), "entry-parsed", G_CALLBACK (entry_parsed), NULL);
+ *
+ * totem_pl_parser_parse_async (pl, "http://example.com/playlist.pls", FALSE, NULL, parse_cb, NULL);
+ * g_object_unref (pl);
+ *
+ * static void
+ * parse_cb (TotemPlParser *parser, GAsyncResult *result, gpointer user_data)
+ * {
+ *	GError *error = NULL;
+ * 	if (totem_pl_parser_parse_finish (parser, result, &error) != TOTEM_PL_PARSER_RESULT_SUCCESS) {
+ * 		g_error ("Playlist parsing failed: %s", error->message);
+ * 		g_error_free (error);
+ * 	}
+ * }
+ *  </programlisting>
+ * </example>
  *
  * <example>
  *  <title>Getting Metadata from Entries</title>
@@ -59,11 +83,14 @@
  *  </programlisting>
  * </example>
  *
+ *
  * <example>
  *  <title>Writing a Playlist</title>
  *  <programlisting>
  * void
- * parser_func (GtkTreeModel *model, GtkTreeIter *iter, gchar **uri, gchar **title, gboolean *custom_title, gpointer user_data)
+ * parser_func (GtkTreeModel *model, GtkTreeIter *iter,
+ * 		gchar **uri, gchar **title, gboolean *custom_title,
+ * 		gpointer user_data)
  * {
  * 	gtk_tree_model_get (model, iter,
  * 		0, uri,
@@ -83,8 +110,10 @@
  * 	tree_model = gtk_list_store_new (3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN);
  * 	populate_model (tree_model);
  *
- * 	if (totem_pl_parser_write (pl, tree_model, parser_func, "/tmp/playlist.pls", TOTEM_PL_PARSER_PLS, NULL, NULL) != TRUE)
+ * 	if (totem_pl_parser_write (pl, tree_model, parser_func, "/tmp/playlist.pls",
+ * 				   TOTEM_PL_PARSER_PLS, NULL, NULL) != TRUE) {
  * 		g_error ("Playlist writing failed.");
+ * 	}
  *
  * 	g_object_unref (tree_model);
  * 	g_object_unref (pl);
@@ -105,8 +134,8 @@
 #include <gobject/gvaluecollector.h>
 #include <gtk/gtk.h>
 
-#ifdef HAVE_CAMEL
-#include <camel/camel-mime-utils.h>
+#ifdef HAVE_GMIME
+#include <gmime/gmime-utils.h>
 #endif
 
 #include "totem-pl-parser.h"
@@ -135,7 +164,7 @@
 typedef const char * (*PlaylistIdenCallback) (const char *data, gsize len);
 
 #ifndef TOTEM_PL_PARSER_MINI
-typedef TotemPlParserResult (*PlaylistCallback) (TotemPlParser *parser, GFile *uri, GFile *base_file, gpointer data);
+typedef TotemPlParserResult (*PlaylistCallback) (TotemPlParser *parser, GFile *uri, GFile *base_file, TotemPlParseData *parse_data, gpointer data);
 #endif
 
 typedef struct {
@@ -222,6 +251,17 @@ static void totem_pl_parser_get_property (GObject *object,
 					  GValue *value,
 					  GParamSpec *pspec);
 
+struct TotemPlParserPrivate {
+	GList *ignore_schemes;
+	GList *ignore_mimetypes;
+	GMutex *ignore_mutex;
+
+	guint recurse : 1;
+	guint debug : 1;
+	guint force : 1;
+	guint disable_unsafe : 1;
+};
+
 enum {
 	PROP_NONE,
 	PROP_RECURSE,
@@ -251,6 +291,8 @@ static void totem_pl_parser_init       (TotemPlParser      *self);
 static void totem_pl_parser_class_init (TotemPlParserClass *klass);
 static gpointer totem_pl_parser_parent_class = NULL;
 
+#define TOTEM_PL_PARSER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), TOTEM_TYPE_PL_PARSER, TotemPlParserPrivate))
+
 GType
 totem_pl_parser_get_type (void)
 {
@@ -268,10 +310,10 @@ totem_pl_parser_get_type (void)
 			0,
 			(GInstanceInitFunc) totem_pl_parser_init,
 		};
-		GType g_define_type_id = g_type_register_static (G_TYPE_OBJECT, "TotemPlParser", &g_define_type_info, 0); 
+		GType g_define_type_id = g_type_register_static (G_TYPE_OBJECT, "TotemPlParser", &g_define_type_info, 0);
 		g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
-	} 
-	return g_define_type_id__volatile; 
+	}
+	return g_define_type_id__volatile;
 }
 
 static void
@@ -281,6 +323,9 @@ totem_pl_parser_class_init (TotemPlParserClass *klass)
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
 	totem_pl_parser_parent_class = g_type_class_peek_parent (klass);
+	if (!g_thread_supported ())
+		g_thread_init (NULL);
+
 	g_type_class_add_private (klass, sizeof (TotemPlParserPrivate));
 
 	object_class->finalize = totem_pl_parser_finalize;
@@ -299,7 +344,7 @@ totem_pl_parser_class_init (TotemPlParserClass *klass)
 					 PROP_RECURSE,
 					 g_param_spec_boolean ("recurse",
 							       "recurse",
-							       "Whether or not to process URIs further", 
+							       "Whether or not to process URIs further",
 							       TRUE,
 							       G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
 
@@ -312,7 +357,7 @@ totem_pl_parser_class_init (TotemPlParserClass *klass)
 					 PROP_DEBUG,
 					 g_param_spec_boolean ("debug",
 							       "debug",
-							       "Whether or not to enable debugging output", 
+							       "Whether or not to enable debugging output",
 							       FALSE,
 							       G_PARAM_READWRITE));
 
@@ -326,7 +371,7 @@ totem_pl_parser_class_init (TotemPlParserClass *klass)
 					 PROP_FORCE,
 					 g_param_spec_boolean ("force",
 							       "force",
-							       "Whether or not to force parsing the file if the playlist looks unsupported", 
+							       "Whether or not to force parsing the file if the playlist looks unsupported",
 							       FALSE,
 							       G_PARAM_READWRITE));
 
@@ -341,7 +386,7 @@ totem_pl_parser_class_init (TotemPlParserClass *klass)
 					 PROP_DISABLE_UNSAFE,
 					 g_param_spec_boolean ("disable-unsafe",
 							       "disable-unsafe",
-							       "Whether or not to disable parsing of unsafe locations", 
+							       "Whether or not to disable parsing of unsafe locations",
 							       FALSE,
 							       G_PARAM_READWRITE));
 
@@ -621,6 +666,26 @@ totem_pl_parser_new (void)
 	return TOTEM_PL_PARSER (g_object_new (TOTEM_TYPE_PL_PARSER, NULL));
 }
 
+typedef struct {
+	TotemPlParser *parser;
+	char *playlist_uri;
+} PlaylistEndedSignalData;
+
+static gboolean
+emit_playlist_ended_signal (PlaylistEndedSignalData *data)
+{
+	g_signal_emit (data->parser,
+		       totem_pl_parser_table_signals[PLAYLIST_ENDED],
+		       0, data->playlist_uri);
+
+	/* Free the data */
+	g_object_unref (data->parser);
+	g_free (data->playlist_uri);
+	g_free (data);
+
+	return FALSE;
+}
+
 /**
  * totem_pl_parser_playlist_end:
  * @parser: a #TotemPlParser
@@ -632,9 +697,13 @@ totem_pl_parser_new (void)
 void
 totem_pl_parser_playlist_end (TotemPlParser *parser, const char *playlist_uri)
 {
-	g_signal_emit (G_OBJECT (parser),
-		       totem_pl_parser_table_signals[PLAYLIST_ENDED],
-		       0, playlist_uri);
+	PlaylistEndedSignalData *data;
+
+	data = g_new (PlaylistEndedSignalData, 1);
+	data->parser = g_object_ref (parser);
+	data->playlist_uri = g_strdup (playlist_uri);
+
+	g_idle_add ((GSourceFunc) emit_playlist_ended_signal, data);
 }
 
 static char *
@@ -699,6 +768,20 @@ my_g_file_info_get_mime_type_with_data (GFile *file, gpointer *data, TotemPlPars
 	return totem_pl_parser_mime_type_from_data (*data, bytes_read);
 }
 
+/**
+ * totem_pl_parser_is_debugging_enabled:
+ * @parser: a #TotemPlParser
+ *
+ * Returns whether debugging is enabled. This is a private method, not exposed by the library.
+ *
+ * Return value: %TRUE if debugging is enabled, %FALSE otherwise
+ **/
+gboolean
+totem_pl_parser_is_debugging_enabled (TotemPlParser *parser)
+{
+	return parser->priv->debug;
+}
+
 /**
  * totem_pl_parser_base_uri:
  * @uri: a URI
@@ -854,6 +937,143 @@ totem_pl_parser_relative (GFile *output, const char *filepath)
 	return retval;
 }
 
+static char *
+relative_uri_remove_query (const char *uri, char **query)
+{
+	char *qmark;
+
+	/* Look for '?' */
+	qmark = strrchr (uri, '?');
+	if (qmark == NULL)
+		return NULL;
+
+	if (query != NULL)
+		*query = g_strdup (qmark);
+	return g_strndup (uri, qmark - uri);
+}
+
+static const char *suffixes[] = {
+	".jsp",
+	".php",
+	".asp"
+};
+
+static gboolean
+is_probably_dir (const char *filename)
+{
+	gboolean ret;
+	char *content_type, *short_name;
+
+	short_name = relative_uri_remove_query (filename, NULL);
+	if (short_name == NULL)
+		short_name = g_strdup (filename);
+	content_type = g_content_type_guess (short_name, NULL, -1, NULL);
+	if (g_content_type_is_unknown (content_type) != FALSE) {
+		guint i;
+		for (i = 0; i < G_N_ELEMENTS (suffixes); i++) {
+			if (g_str_has_suffix (short_name, suffixes[i]) != FALSE) {
+				g_free (content_type);
+				g_free (short_name);
+				return FALSE;
+			}
+		}
+		ret = TRUE;
+	} else {
+		ret = FALSE;
+	}
+	g_free (content_type);
+	g_free (short_name);
+
+	return ret;
+}
+
+char *
+totem_pl_parser_resolve_uri (GFile *base_gfile,
+			     const char *relative_uri)
+{
+	char *uri, *scheme, *query, *new_relative_uri, *base_uri;
+	GFile *base_parent_gfile, *resolved_gfile;
+
+	if (relative_uri == NULL) {
+		if (base_gfile == NULL)
+			return NULL;
+		return g_file_get_uri (base_gfile);
+	}
+
+	if (base_gfile == NULL)
+		return g_strdup (relative_uri);
+
+	/* If |relative_uri| has a scheme, it's a full URI, just return it */
+	scheme = g_uri_parse_scheme (relative_uri);
+	if (scheme != NULL) {
+		g_free (scheme);
+		return g_strdup (relative_uri);
+	}
+
+	/* Check whether we need to get the parent for the base or not */
+	base_uri = g_file_get_path (base_gfile);
+	if (base_uri == NULL)
+		base_uri = g_file_get_uri (base_gfile);
+	if (is_probably_dir (base_uri) == FALSE)
+		base_parent_gfile = g_file_get_parent (base_gfile);
+	else
+		base_parent_gfile = g_object_ref (base_gfile);
+	g_free (base_uri);
+
+	if (base_parent_gfile == NULL) {
+		resolved_gfile = g_file_resolve_relative_path (base_gfile, relative_uri);
+		uri = g_file_get_uri (resolved_gfile);
+		g_object_unref (resolved_gfile);
+		return uri;
+	}
+
+	/* Remove the query portion of the URI, to transplant it again
+	 * if there is any */
+	query = NULL;
+	new_relative_uri = relative_uri_remove_query (relative_uri, &query);
+
+	if (new_relative_uri) {
+		char *tmpuri;
+
+		resolved_gfile = g_file_resolve_relative_path (base_parent_gfile, new_relative_uri);
+		g_object_unref (base_parent_gfile);
+		if (!resolved_gfile) {
+			char *base_uri;
+			base_uri = g_file_get_uri (base_gfile);
+			g_warning ("Failed to resolve relative URI '%s' against base '%s'\n", relative_uri, base_uri);
+			g_free (base_uri);
+			g_free (new_relative_uri);
+			g_free (query);
+			return NULL;
+		}
+
+		tmpuri = g_file_get_uri (resolved_gfile);
+		g_object_unref (resolved_gfile);
+		uri = g_strdup_printf ("%s%s", tmpuri, query);
+
+		g_free (tmpuri);
+		g_free (new_relative_uri);
+		g_free (query);
+
+		return uri;
+	} else {
+		resolved_gfile = g_file_resolve_relative_path (base_parent_gfile, relative_uri);
+		g_object_unref (base_parent_gfile);
+		if (!resolved_gfile) {
+			char *base_uri;
+			base_uri = g_file_get_uri (base_gfile);
+			g_warning ("Failed to resolve relative URI '%s' against base '%s'\n", relative_uri, base_uri);
+			g_free (base_uri);
+			return NULL;
+		}
+
+		uri = g_file_get_uri (resolved_gfile);
+		g_object_unref (resolved_gfile);
+
+		return uri;
+	}
+}
+
 #ifndef TOTEM_PL_PARSER_MINI
 /**
  * totem_pl_parser_write_with_title:
@@ -1007,7 +1227,6 @@ totem_pl_parser_read_ini_line_int (char **lines, const char *key)
  * totem_pl_parser_read_ini_line_string_with_sep:
  * @lines: a NULL-terminated array of INI lines to read
  * @key: the key to match
- * @dos_mode: %TRUE if the returned string should end in \r\0, instead of \n\0
  * @sep: the key-value separator
  *
  * Returns the first string value case-insensitively matching the specified
@@ -1018,7 +1237,7 @@ totem_pl_parser_read_ini_line_int (char **lines, const char *key)
  **/
 char*
 totem_pl_parser_read_ini_line_string_with_sep (char **lines, const char *key,
-		gboolean dos_mode, const char *sep)
+		const char *sep)
 {
 	char *retval = NULL;
 	int i;
@@ -1034,7 +1253,6 @@ totem_pl_parser_read_ini_line_string_with_sep (char **lines, const char *key,
 
 		if (g_ascii_strncasecmp (line, key, strlen (key)) == 0) {
 			char **bits;
-			glong len;
 
 			bits = g_strsplit (line, sep, 2);
 			if (bits[0] == NULL || bits [1] == NULL) {
@@ -1043,12 +1261,6 @@ totem_pl_parser_read_ini_line_string_with_sep (char **lines, const char *key,
 			}
 
 			retval = g_strdup (bits[1]);
-			len = strlen (retval);
-			if (dos_mode && len >= 2 && retval[len-2] == '\r') {
-				retval[len-2] = '\n';
-				retval[len-1] = '\0';
-			}
-
 			g_strfreev (bits);
 		}
 	}
@@ -1060,7 +1272,6 @@ totem_pl_parser_read_ini_line_string_with_sep (char **lines, const char *key,
  * totem_pl_parser_read_ini_line_string:
  * @lines: a NULL-terminated array of INI lines to read
  * @key: the key to match
- * @dos_mode: %TRUE if the returned string should end in \r\0, instead of \n\0
  *
  * Returns the first string value case-insensitively matching the
  * specified key. The parser ignores leading whitespace on lines.
@@ -1068,34 +1279,58 @@ totem_pl_parser_read_ini_line_string_with_sep (char **lines, const char *key,
  * Return value: a newly-allocated string value, or %NULL
  **/
 char*
-totem_pl_parser_read_ini_line_string (char **lines, const char *key, gboolean dos_mode)
+totem_pl_parser_read_ini_line_string (char **lines, const char *key)
 {
-	return totem_pl_parser_read_ini_line_string_with_sep (lines, key, dos_mode, "=");
+	return totem_pl_parser_read_ini_line_string_with_sep (lines, key, "=");
 }
 
 static void
 totem_pl_parser_init (TotemPlParser *parser)
 {
 	parser->priv = G_TYPE_INSTANCE_GET_PRIVATE (parser, TOTEM_TYPE_PL_PARSER, TotemPlParserPrivate);
+	parser->priv->ignore_mutex = g_mutex_new ();
 }
 
 static void
 totem_pl_parser_finalize (GObject *object)
 {
-	TotemPlParser *parser = TOTEM_PL_PARSER (object);
+	TotemPlParserPrivate *priv = TOTEM_PL_PARSER (object)->priv;
 
 	g_return_if_fail (object != NULL);
-	g_return_if_fail (parser->priv != NULL);
+	g_return_if_fail (priv != NULL);
 
-	g_list_foreach (parser->priv->ignore_schemes, (GFunc) g_free, NULL);
-	g_list_free (parser->priv->ignore_schemes);
+	g_list_foreach (priv->ignore_schemes, (GFunc) g_free, NULL);
+	g_list_free (priv->ignore_schemes);
 
-	g_list_foreach (parser->priv->ignore_mimetypes, (GFunc) g_free, NULL);
-	g_list_free (parser->priv->ignore_mimetypes);
+	g_list_foreach (priv->ignore_mimetypes, (GFunc) g_free, NULL);
+	g_list_free (priv->ignore_mimetypes);
+
+	g_mutex_free (priv->ignore_mutex);
 
 	G_OBJECT_CLASS (totem_pl_parser_parent_class)->finalize (object);
 }
 
+typedef struct {
+	TotemPlParser *parser;
+	guint signal_id;
+	char *uri;
+	GHashTable *metadata;
+} EntryParsedSignalData;
+
+static gboolean
+emit_entry_parsed_signal (EntryParsedSignalData *data)
+{
+	g_signal_emit (data->parser, data->signal_id, 0, data->uri, data->metadata);
+
+	/* Free the data */
+	g_object_unref (data->parser);
+	g_free (data->uri);
+	g_hash_table_unref (data->metadata);
+	g_free (data);
+
+	return FALSE;
+}
+
 static void
 totem_pl_parser_add_uri_valist (TotemPlParser *parser,
 				const gchar *first_property_name,
@@ -1208,15 +1443,21 @@ totem_pl_parser_add_uri_valist (TotemPlParser *parser,
 	}
 
 	if (g_hash_table_size (metadata) > 0 || uri != NULL) {
-		if (is_playlist == FALSE) {
-			g_signal_emit (G_OBJECT (parser),
-				       totem_pl_parser_table_signals[ENTRY_PARSED],
-				       0, uri, metadata);
-		} else {
-			g_signal_emit (G_OBJECT (parser),
-				       totem_pl_parser_table_signals[PLAYLIST_STARTED],
-				       0, uri, metadata);
-		}
+		EntryParsedSignalData *data;
+
+		/* Make sure to emit the signals asynchronously, as we could be in the main loop
+		 * *or* a worker thread at this point. */
+		data = g_new (EntryParsedSignalData, 1);
+		data->parser = g_object_ref (parser);
+		data->uri = g_strdup (uri);
+		data->metadata = g_hash_table_ref (metadata);
+
+		if (is_playlist == FALSE)
+			data->signal_id = totem_pl_parser_table_signals[ENTRY_PARSED];
+		else
+			data->signal_id = totem_pl_parser_table_signals[PLAYLIST_STARTED];
+
+		g_idle_add ((GSourceFunc) emit_entry_parsed_signal, data);
 	}
 
 	g_hash_table_unref (metadata);
@@ -1295,15 +1536,23 @@ totem_pl_parser_scheme_is_ignored (TotemPlParser *parser, GFile *uri)
 {
 	GList *l;
 
-	if (parser->priv->ignore_schemes == NULL)
+	g_mutex_lock (parser->priv->ignore_mutex);
+
+	if (parser->priv->ignore_schemes == NULL) {
+		g_mutex_unlock (parser->priv->ignore_mutex);
 		return FALSE;
+	}
 
 	for (l = parser->priv->ignore_schemes; l != NULL; l = l->next) {
 		const char *scheme = l->data;
-		if (g_file_has_uri_scheme (uri, scheme) != FALSE)
+		if (g_file_has_uri_scheme (uri, scheme) != FALSE) {
+			g_mutex_unlock (parser->priv->ignore_mutex);
 			return TRUE;
+		}
 	}
 
+	g_mutex_unlock (parser->priv->ignore_mutex);
+
 	return FALSE;
 }
 
@@ -1313,16 +1562,24 @@ totem_pl_parser_mimetype_is_ignored (TotemPlParser *parser,
 {
 	GList *l;
 
-	if (parser->priv->ignore_mimetypes == NULL)
+	g_mutex_lock (parser->priv->ignore_mutex);
+
+	if (parser->priv->ignore_mimetypes == NULL) {
+		g_mutex_unlock (parser->priv->ignore_mutex);
 		return FALSE;
+	}
 
 	for (l = parser->priv->ignore_mimetypes; l != NULL; l = l->next)
 	{
 		const char *item = l->data;
-		if (strcmp (mimetype, item) == 0)
+		if (strcmp (mimetype, item) == 0) {
+			g_mutex_unlock (parser->priv->ignore_mutex);
 			return TRUE;
+		}
 	}
 
+	g_mutex_unlock (parser->priv->ignore_mutex);
+
 	return FALSE;
 }
 
@@ -1390,7 +1647,7 @@ totem_pl_parser_ignore (TotemPlParser *parser, const char *uri)
  * Removes HTML comments from a string representing the contents of an XML file.
  * The function modifies the string in place.
  */
-void
+static void
 totem_pl_parser_cleanup_xml (char *contents)
 {
 	char *needle;
@@ -1410,6 +1667,53 @@ totem_pl_parser_cleanup_xml (char *contents)
 	}
 }
 
+xml_node_t *
+totem_pl_parser_parse_xml_relaxed (char *contents,
+				   gsize size)
+{
+	xml_node_t* doc, *node;
+	char *encoding, *new_contents;
+	gsize new_size;
+
+	totem_pl_parser_cleanup_xml (contents);
+	xml_parser_init (contents, size, XML_PARSER_CASE_INSENSITIVE);
+	if (xml_parser_build_tree_with_options (&doc, XML_PARSER_RELAXED | XML_PARSER_MULTI_TEXT) < 0)
+		return NULL;
+
+	encoding = NULL;
+	for (node = doc; node != NULL; node = node->next) {
+		if (node->name == NULL || g_str_equal (node->name, "?XML") == FALSE)
+			continue;
+		encoding = g_strdup (xml_parser_get_property (node, "ENCODING"));
+		break;
+	}
+
+	if (encoding == NULL || g_str_equal (encoding, "UTF-8") != FALSE) {
+		g_free (encoding);
+		return doc;
+	}
+
+	xml_parser_free_tree (doc);
+
+	new_contents = g_convert (contents, size, "UTF-8", encoding, NULL, &new_size, NULL);
+	if (new_contents == NULL) {
+		g_warning ("Failed to convert XML data to UTF-8");
+		g_free (encoding);
+		return NULL;
+	}
+	g_free (encoding);
+
+	xml_parser_init (new_contents, new_size, XML_PARSER_CASE_INSENSITIVE);
+	if (xml_parser_build_tree_with_options (&doc, XML_PARSER_RELAXED | XML_PARSER_MULTI_TEXT) < 0) {
+		g_free (new_contents);
+		return NULL;
+	}
+
+	g_free (new_contents);
+
+	return doc;
+}
+
 static gboolean
 totem_pl_parser_ignore_from_mimetype (TotemPlParser *parser, const char *mimetype)
 {
@@ -1431,7 +1735,8 @@ totem_pl_parser_ignore_from_mimetype (TotemPlParser *parser, const char *mimetyp
 TotemPlParserResult
 totem_pl_parser_parse_internal (TotemPlParser *parser,
 				GFile *file,
-				GFile *base_file)
+				GFile *base_file,
+				TotemPlParseData *parse_data)
 {
 	char *mimetype;
 	guint i;
@@ -1439,7 +1744,7 @@ totem_pl_parser_parse_internal (TotemPlParser *parser,
 	TotemPlParserResult ret = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 	gboolean found = FALSE;
 
-	if (parser->priv->recurse_level > RECURSE_LEVEL_MAX)
+	if (parse_data->recurse_level > RECURSE_LEVEL_MAX)
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	if (g_file_has_uri_scheme (file, "mms") != FALSE
@@ -1458,23 +1763,23 @@ totem_pl_parser_parse_internal (TotemPlParser *parser,
 	    || g_file_has_uri_scheme (file, "feed") != FALSE
 	    || g_file_has_uri_scheme (file, "zcast") != FALSE) {
 		DEBUG(file, g_print ("URI '%s' is getting special cased for ITPC/FEED/ZCAST parsing\n", uri));
-		return totem_pl_parser_add_itpc (parser, file, base_file, NULL);
+		return totem_pl_parser_add_itpc (parser, file, base_file, parse_data, NULL);
 	}
 	if (g_file_has_uri_scheme (file, "zune") != FALSE) {
 		DEBUG(file, g_print ("URI '%s' is getting special cased for ZUNE parsing\n", uri));
-		return totem_pl_parser_add_zune (parser, file, base_file, NULL);
+		return totem_pl_parser_add_zune (parser, file, base_file, parse_data, NULL);
 	}
 	/* Try itms Podcast references, see itunes.py in PenguinTV */
 	if (totem_pl_parser_is_itms_feed (file) != FALSE) {
-	    	DEBUG(file, g_print ("URI '%s' is getting special cased for ITMS parsing\n", uri));
-	    	return totem_pl_parser_add_itms (parser, file, NULL, NULL);
+		DEBUG(file, g_print ("URI '%s' is getting special cased for ITMS parsing\n", uri));
+		return totem_pl_parser_add_itms (parser, file, NULL, parse_data, NULL);
 	}
 
-	if (!parser->priv->recurse && parser->priv->recurse_level > 0)
+	if (!parse_data->recurse && parse_data->recurse_level > 0)
 		return TOTEM_PL_PARSER_RESULT_UNHANDLED;
 
 	/* In force mode we want to get the data */
-	if (parser->priv->force != FALSE) {
+	if (parse_data->force != FALSE) {
 		mimetype = my_g_file_info_get_mime_type_with_data (file, &data, parser);
 	} else {
 		char *uri;
@@ -1521,7 +1826,7 @@ totem_pl_parser_parse_internal (TotemPlParser *parser,
 
 	/* If we're at the top-level of the parsing, try to get more
 	 * data from the playlist parser */
-	if (strcmp (mimetype, AUDIO_MPEG_TYPE) == 0 && parser->priv->recurse_level == 0 && data == NULL) {
+	if (strcmp (mimetype, AUDIO_MPEG_TYPE) == 0 && parse_data->recurse_level == 0 && data == NULL) {
 		char *tmp;
 		tmp = my_g_file_info_get_mime_type_with_data (file, &data, parser);
 		if (tmp != NULL) {
@@ -1537,13 +1842,13 @@ totem_pl_parser_parse_internal (TotemPlParser *parser,
 		return TOTEM_PL_PARSER_RESULT_IGNORED;
 	}
 
-	if (parser->priv->recurse || parser->priv->recurse_level == 0) {
-		parser->priv->recurse_level++;
+	if (parse_data->recurse || parse_data->recurse_level == 0) {
+		parse_data->recurse_level++;
 
 		for (i = 0; i < G_N_ELEMENTS(special_types); i++) {
 			if (strcmp (special_types[i].mimetype, mimetype) == 0) {
 				DEBUG(file, g_print ("URI '%s' is special type '%s'\n", uri, mimetype));
-				if (parser->priv->disable_unsafe != FALSE && special_types[i].unsafe != FALSE) {
+				if (parse_data->disable_unsafe != FALSE && special_types[i].unsafe != FALSE) {
 					DEBUG(file, g_print ("URI '%s' is unsafe so was ignored\n", uri));
 					g_free (mimetype);
 					g_free (data);
@@ -1555,7 +1860,7 @@ totem_pl_parser_parse_internal (TotemPlParser *parser,
 					base_file = g_object_ref (base_file);
 
 				DEBUG (file, g_print ("Using %s function for '%s'\n", special_types[i].mimetype, uri));
-				ret = (* special_types[i].func) (parser, file, base_file, data);
+				ret = (* special_types[i].func) (parser, file, base_file, parse_data, data);
 
 				if (base_file != NULL)
 					g_object_unref (base_file);
@@ -1583,7 +1888,7 @@ totem_pl_parser_parse_internal (TotemPlParser *parser,
 				else
 					base_file = g_object_ref (base_file);
 
-				ret = (* dual_types[i].func) (parser, file, base_file ? base_file : file, data);
+				ret = (* dual_types[i].func) (parser, file, base_file ? base_file : file, parse_data, data);
 
 				if (base_file != NULL)
 					g_object_unref (base_file);
@@ -1595,7 +1900,7 @@ totem_pl_parser_parse_internal (TotemPlParser *parser,
 
 		g_free (data);
 
-		parser->priv->recurse_level--;
+		parse_data->recurse_level--;
 	}
 
 	if (ret == TOTEM_PL_PARSER_RESULT_SUCCESS) {
@@ -1609,7 +1914,7 @@ totem_pl_parser_parse_internal (TotemPlParser *parser,
 	}
 	g_free (mimetype);
 
-	if (ret != TOTEM_PL_PARSER_RESULT_SUCCESS && parser->priv->fallback) {
+	if (ret != TOTEM_PL_PARSER_RESULT_SUCCESS && parse_data->fallback) {
 		totem_pl_parser_add_one_file (parser, file, NULL);
 		return TOTEM_PL_PARSER_RESULT_SUCCESS;
 	}
@@ -1617,6 +1922,81 @@ totem_pl_parser_parse_internal (TotemPlParser *parser,
 	return ret;
 }
 
+typedef struct {
+	char *uri;
+	char *base;
+	gboolean fallback;
+} ParseAsyncData;
+
+static void
+parse_async_data_free (ParseAsyncData *data)
+{
+	g_free (data->uri);
+	g_free (data->base);
+	g_slice_free (ParseAsyncData, data);
+}
+
+static void
+parse_thread (GSimpleAsyncResult *result, GObject *object, GCancellable *cancellable)
+{
+	TotemPlParserResult parse_result;
+	GError *error = NULL;
+	ParseAsyncData *data = g_simple_async_result_get_op_res_gpointer (result);
+
+	/* Check to see if it's been cancelled already */
+	if (g_cancellable_set_error_if_cancelled (cancellable, &error) == TRUE) {
+		g_simple_async_result_set_from_error (result, error);
+		g_simple_async_result_set_op_res_gpointer (result, GUINT_TO_POINTER (TOTEM_PL_PARSER_RESULT_CANCELLED), NULL);
+		g_error_free (error);
+		return;
+	}
+
+	/* Parse and return */
+	parse_result = totem_pl_parser_parse_with_base (TOTEM_PL_PARSER (object), data->uri, data->base, data->fallback);
+	g_simple_async_result_set_op_res_gpointer (result, GUINT_TO_POINTER (parse_result), NULL);
+}
+
+/**
+ * totem_pl_parser_parse_with_base_async:
+ * @parser: a #TotemPlParser
+ * @uri: the URI of the playlist to parse
+ * @base: the base path for relative filenames
+ * @fallback: %TRUE if the parser should add the playlist URI to the
+ * end of the playlist on parse failure
+ * @cancellable: optional #GCancellable object, or %NULL
+ * @callback: a #GAsyncReadyCallback to call when parsing is finished
+ * @user_data: data to pass to the @callback function
+ *
+ * Starts asynchronous parsing of a playlist given by the absolute URI @uri, using @base to resolve relative paths where appropriate.
+ * @self and @uri are both reffed/copied when this function is called, so can safely be freed after this function returns.
+ *
+ * For more details, see totem_pl_parser_parse_with_base(), which is the synchronous version of this function.
+ *
+ * When the operation is finished, @callback will be called. You can then call totem_pl_parser_parse_finish()
+ * to get the results of the operation.
+ **/
+void
+totem_pl_parser_parse_with_base_async (TotemPlParser *parser, const char *uri, const char *base, gboolean fallback,
+				       GCancellable *cancellable, GAsyncReadyCallback callback, gpointer user_data)
+{
+	GSimpleAsyncResult *result;
+	ParseAsyncData *data;
+
+	g_return_if_fail (TOTEM_IS_PL_PARSER (parser));
+	g_return_if_fail (uri != NULL);
+	g_return_if_fail (strstr (uri, "://") != NULL);
+
+	data = g_slice_new (ParseAsyncData);
+	data->uri = g_strdup (uri);
+	data->base = g_strdup (base);
+	data->fallback = fallback;
+
+	result = g_simple_async_result_new (G_OBJECT (parser), callback, user_data, totem_pl_parser_parse_with_base_async);
+	g_simple_async_result_set_op_res_gpointer (result, data, (GDestroyNotify) parse_async_data_free);
+	g_simple_async_result_run_in_thread (result, (GSimpleAsyncThreadFunc) parse_thread, G_PRIORITY_DEFAULT, cancellable);
+	g_object_unref (result);
+}
+
 /**
  * totem_pl_parser_parse_with_base:
  * @parser: a #TotemPlParser
@@ -1636,6 +2016,7 @@ totem_pl_parser_parse_with_base (TotemPlParser *parser, const char *uri,
 {
 	GFile *file, *base_file;
 	TotemPlParserResult retval;
+	TotemPlParseData data;
 
 	g_return_val_if_fail (TOTEM_IS_PL_PARSER (parser), TOTEM_PL_PARSER_RESULT_UNHANDLED);
 	g_return_val_if_fail (uri != NULL, TOTEM_PL_PARSER_RESULT_UNHANDLED);
@@ -1650,11 +2031,16 @@ totem_pl_parser_parse_with_base (TotemPlParser *parser, const char *uri,
 		return TOTEM_PL_PARSER_RESULT_UNHANDLED;
 	}
 
-	parser->priv->recurse_level = 0;
-	parser->priv->fallback = fallback != FALSE;
+	/* Use a struct to store copies of the options as set for this parse operation */
+	data.recurse_level = 0;
+	data.fallback = fallback;
+	data.recurse = parser->priv->recurse;
+	data.force = parser->priv->force;
+	data.disable_unsafe = parser->priv->disable_unsafe;
+
 	if (base != NULL)
 		base_file = g_file_new_for_uri (base);
-	retval = totem_pl_parser_parse_internal (parser, file, base_file);
+	retval = totem_pl_parser_parse_internal (parser, file, base_file, &data);
 
 	g_object_unref (file);
 	if (base_file != NULL)
@@ -1663,6 +2049,60 @@ totem_pl_parser_parse_with_base (TotemPlParser *parser, const char *uri,
 	return retval;
 }
 
+/**
+ * totem_pl_parser_parse_async:
+ * @parser: a #TotemPlParser
+ * @uri: the URI of the playlist to parse
+ * @fallback: %TRUE if the parser should add the playlist URI to the
+ * end of the playlist on parse failure
+ * @cancellable: optional #GCancellable object, or %NULL
+ * @callback: a #GAsyncReadyCallback to call when parsing is finished
+ * @user_data: data to pass to the @callback function
+ *
+ * Starts asynchronous parsing of a playlist given by the absolute URI @uri. @self and @uri are both reffed/copied
+ * when this function is called, so can safely be freed after this function returns.
+ *
+ * For more details, see totem_pl_parser_parse(), which is the synchronous version of this function.
+ *
+ * When the operation is finished, @callback will be called. You can then call totem_pl_parser_parse_finish()
+ * to get the results of the operation.
+ **/
+void
+totem_pl_parser_parse_async (TotemPlParser *parser, const char *uri, gboolean fallback,
+			     GCancellable *cancellable, GAsyncReadyCallback callback, gpointer user_data)
+{
+	totem_pl_parser_parse_with_base_async (parser, uri, NULL, fallback, cancellable, callback, user_data);
+}
+
+/**
+ * totem_pl_parser_parse_finish:
+ * @parser: a #TotemPlParser
+ * @async_result: a #GAsyncResult
+ * @error: a #GError, or %NULL
+ *
+ * Finishes an asynchronous playlist parsing operation started with totem_pl_parser_parse_async()
+ * or totem_pl_parser_parse_with_base_async().
+ *
+ * If parsing of the playlist is cancelled part-way through, %TOTEM_PL_PARSER_RESULT_CANCELLED is returned when
+ * this function is called.
+ *
+ * Return value: a #TotemPlParserResult
+ **/
+TotemPlParserResult
+totem_pl_parser_parse_finish (TotemPlParser *parser, GAsyncResult *async_result, GError **error)
+{
+	GSimpleAsyncResult *result = G_SIMPLE_ASYNC_RESULT (async_result);
+
+	g_return_val_if_fail (TOTEM_IS_PL_PARSER (parser), FALSE);
+	g_return_val_if_fail (G_IS_ASYNC_RESULT (async_result), FALSE);
+
+	g_warn_if_fail (g_simple_async_result_get_source_tag (result) == totem_pl_parser_parse_with_base_async);
+
+	/* Propagate any errors which were caught and return the result; otherwise just return the result */
+	g_simple_async_result_propagate_error (result, error);
+	return GPOINTER_TO_UINT (g_simple_async_result_get_op_res_gpointer (result));
+}
+
 /**
  * totem_pl_parser_parse:
  * @parser: a #TotemPlParser
@@ -1670,7 +2110,11 @@ totem_pl_parser_parse_with_base (TotemPlParser *parser, const char *uri,
  * @fallback: %TRUE if the parser should add the playlist URI to the
  * end of the playlist on parse failure
  *
- * Parses a playlist given by the absolute URI @uri.
+ * Parses a playlist given by the absolute URI @uri. This method is
+ * synchronous, and will block on (e.g.) network requests to slow
+ * servers. totem_pl_parser_parse_async() is recommended instead.
+ *
+ * Return values are as totem_pl_parser_parse_with_base().
  *
  * Return value: a #TotemPlParserResult
  **/
@@ -1697,11 +2141,15 @@ totem_pl_parser_add_ignored_scheme (TotemPlParser *parser,
 
 	g_return_if_fail (TOTEM_IS_PL_PARSER (parser));
 
+	g_mutex_lock (parser->priv->ignore_mutex);
+
 	s = g_strdup (scheme);
 	if (s[strlen (s) - 1] == ':')
 		s[strlen (s) - 1] = '\0';
 	parser->priv->ignore_schemes = g_list_prepend
 		(parser->priv->ignore_schemes, s);
+
+	g_mutex_unlock (parser->priv->ignore_mutex);
 }
 
 /**
@@ -1718,8 +2166,12 @@ totem_pl_parser_add_ignored_mimetype (TotemPlParser *parser,
 {
 	g_return_if_fail (TOTEM_IS_PL_PARSER (parser));
 
+	g_mutex_lock (parser->priv->ignore_mutex);
+
 	parser->priv->ignore_mimetypes = g_list_prepend
 		(parser->priv->ignore_mimetypes, g_strdup (mimetype));
+
+	g_mutex_unlock (parser->priv->ignore_mutex);
 }
 
 /**
@@ -1800,7 +2252,7 @@ totem_pl_parser_parse_duration (const char *duration, gboolean debug)
 guint64
 totem_pl_parser_parse_date (const char *date_str, gboolean debug)
 {
-#ifdef HAVE_CAMEL
+#ifdef HAVE_GMIME
 	GTimeVal val;
 
 	g_return_val_if_fail (date_str != NULL, -1);
@@ -1813,10 +2265,10 @@ totem_pl_parser_parse_date (const char *date_str, gboolean debug)
 	}
 	D(g_message ("Failed to parse duration '%s' using the ISO8601 parser", date_str));
 	/* Fall back to RFC 2822 date parsing */
-	return camel_header_decode_date (date_str, NULL);
+	return g_mime_utils_header_decode_date (date_str, NULL);
 #else
-	WARN_NO_CAMEL;
-#endif /* HAVE_CAMEL */
+	WARN_NO_GMIME;
+#endif /* HAVE_GMIME */
 }
 #endif /* !TOTEM_PL_PARSER_MINI */
 
@@ -1967,7 +2419,7 @@ totem_pl_parser_metadata_get_type (void)
 {
 	static volatile gsize g_define_type_id__volatile = 0;
 	if (g_once_init_enter (&g_define_type_id__volatile))
-	{ 
+	{
 		GType g_define_type_id = g_boxed_type_register_static (
 		    g_intern_static_string ("TotemPlParserMetadata"),
 		    (GBoxedCopyFunc) g_hash_table_ref,
